/**
 * @file
 *
 * This file contains the functions for all the implemented mapping
 * heuristics.
 * The mapping heuristics generate solutions based on the
 * task graph of the group of tasks and the network 
 * performance model of the network.  These heuristics
 * generate the best mapping solution (best_mg) in
 * gs_smart_mapping_solution.h
 *
 *
 */



#include "config.h"

#ifdef GS_SMART_GRIDSOLVE
#include <math.h>
#include "gs_smart_task_graph.h"
#include "gs_smart_mapping_graph.h"
#include "gs_smart_mapping_heuristics.h"
#include "gs_smart_mapping_matrix_func.h"

/*
 * The list of names of implemented mapping heuristics
 */

static int nb_mappers=4;
static char mapping_names[4][15] ={"ex_map", "basic_map", "random_map", "rwalk_map"};


/*
 * The name of the mapping heuristic called
 */
static char * mapper_type;


/*
 *  This function invokes the called mapping heuristic
 *  @param tg - (in) Task graph of a the mapped group of tasks
 *  @param netpm -- (in) Network performance model
 *  @param best_mg -- (out) Mapping solution generated by the mapping heuristic 
 *  @param mapper_num -- (in) The position of mapping solution to invoke in mapping_names array.
 */

int gs_smart_exec_mapper(gs_smart_tg * tg, gs_smart_netpm * netpm, gs_smart_mg ** _best_mg, int mapper_num){
  LOGPRINTF("Mapping group of tasks with mapping heuristic %s\n", mapping_names[mapper_num]);
   gs_smart_mg * best_mg;
  if(mapper_num==0){
    if(gs_smart_ex_map(tg, netpm, &best_mg)<0){
        ERRPRINTF("SMART : Error performing exhaustive mapping heuristic\n");
        return -1;
    }
  }
  else if(mapper_num==1){
    if(gs_smart_basic_map(tg, netpm, &best_mg)<0){
        ERRPRINTF("SMART : Error performing basic mapping heuristic\n");
        return -1;
    }
  }
  else if(mapper_num==2){
    if(gs_smart_random_map(tg, netpm, &best_mg)<0){
        ERRPRINTF("SMART : Error performing random mapping heuristic\n");
        return -1;
    }
  }
  else if(mapper_num==3){
    if(gs_smart_rwalk_map(tg, netpm, &best_mg)<0){
        ERRPRINTF("SMART : Error performing random walk mapping heuristic\n");
        return -1;
    }
  }
  *_best_mg=best_mg;
  return 0;
}







/*
 *  The exhaustive search mapping heuristic is an iterative mapping 
 *  heuristic.
 *  This mapping heuristic iterates through every possible mapping 
 *  solution and returns the mapping solution with fastest time.
 *
 *  This heuristic can be time-consuming when there are a large 
 *  number of solutions.  The number of solutions for a given 
 *  set of tasks and server is equal to num_server to the power
 *  of num_tasks, which can be a significant number.
 *  This can be expensive as it roughly takes one second to time 
 *  10,000 solutions.
 *
 *  @param tg - (in) Task graph of a the mapped group of tasks
 *  @param netpm -- (in) Network performance model
 *  @param best_mg -- (out) Mapping solution generated by mapping heuristic 
 *
 */



int gs_smart_ex_map(gs_smart_tg * tg, gs_smart_netpm * netpm, gs_smart_mg ** _best_mg){

  gs_smart_mg * this_mg;
  gs_smart_mg * best_mg=NULL;
  int ** mapping_matrix;
  int * mapping_vec;  
  int nb_rem_tasks, nb_servers;
  int i, j;
  int total_permutations, this_per;
  double best_time=0.0;
  double this_time=0.0;
  if(tg->nb_rem_nodes<=0){
    ERRPRINTF("SMART: Error doing exhaustive search, no remote nodes\n");
    return -1;
  }

  nb_servers=netpm->nb_nodes;
  nb_rem_tasks=tg->nb_rem_nodes;
  mapping_matrix = (int **)calloc( nb_servers, sizeof(int *));

  if(!mapping_matrix) return -1;
  
  for (i = 0; i <nb_servers; i++) {
    mapping_matrix[i] = (int *)calloc(nb_rem_tasks, sizeof(int));
    if(!mapping_matrix[i]) return -1;
  }

  for( i = 0; i < nb_servers; i++){
    for( j = 0; j < nb_rem_tasks; j++){
      mapping_matrix[i][j] = 0;
    }
  }

  mapping_vec = (int *)calloc(nb_rem_tasks, sizeof(int));
  if(!mapping_vec){
    ERRPRINTF("SMART: Error allocating to mapping vector\n");
    return -1;
  } 

  total_permutations=(int)(pow((double)nb_servers,(double)nb_rem_tasks));


  int num_valid_graphs=0;
 
  for(this_per=0;this_per<total_permutations;this_per++){
    this_mg = (gs_smart_mg * )calloc(1, sizeof(gs_smart_mg));
    if(gs_smart_zero_mapping_matrix(mapping_matrix, nb_servers, nb_rem_tasks)<0){
      ERRPRINTF("SMART: Error resetting mapping matrix on mapping solution %d\n", this_per);
      return -1;
    }

    
    if(gs_smart_next_permutation(mapping_matrix, this_per, nb_servers, nb_rem_tasks)<0){
      ERRPRINTF("SMART: Error shifting matrix to next permuation on mapping solution %d\n", this_per);
      return -1;
    }
    if(gs_smart_matrix_to_vec(mapping_matrix, nb_servers, nb_rem_tasks, mapping_vec)<0){
      ERRPRINTF("SMART: Error converting mapping matrix to a vector\n");
      return -1;
    }
    int valid_graph;
    if(gs_smart_generate_mapping_graph(tg, netpm, mapping_vec, this_mg, &valid_graph)<0){
      ERRPRINTF("SMART: Error generating mapping graph\n");
      return -1;
    }

   this_time=this_mg->total_time;
    if(valid_graph){ 
      if(num_valid_graphs==0){
        best_time=this_time;
        best_mg=this_mg;
      }
      else{
        if(this_time<best_time){
          best_time=this_time;
          if(gs_smart_free_mg(best_mg)<0){
            ERRPRINTF("SMART : Error freeing mapping graph\n");
            return -1;
          }
          best_mg=this_mg;
        }
        else{
          if(gs_smart_free_mg(this_mg)<0){
            ERRPRINTF("SMART : Error freeing mapping graph\n");
            return -1;
          }
        }

      }
      num_valid_graphs++;
     }
  }
  if(num_valid_graphs==0){
    ERRPRINTF("SMART : No mapping solutions found for this group of tasks\n");
    return -1;

  }
  if(gs_smart_mg_create_argument_file_names(best_mg)<0){
    ERRPRINTF("SMART : Error creating mapping graph argument file names\n");
    return -1;
  }
  *_best_mg=best_mg;
  return 0;
}







/*
 *  The basic mapping heuristic is a non-iterative mapping heuristic.
 *  It maps each task one at a time. It finds the fastest smart enabled
 *  server that can execute each task.  When a nonblocking task gets 
 *  assigned to a server, the server's workload is increased. 
 *  This provides a basic method for load balancing non-blocking tasks.
 *  If there are no smart servers that can execute a specific task
 *  then this task will be assigned a server that is not smart enabled.
 *  Once the tasks have been assigned to servers, the mapping graph is
 *  generated with the best communication scheme, based on the dependencies
 *  between tasks and whether the servers assigned are smart servers or not.
 *
 *  This is a basic heuristic which generates a mapping solution quickly 
 *  and provides speed up due to direct communication between servers.
 *  However since it maps tasks individually it does not fully exploit the 
 *  potential of mapping tasks collectively such as improved load balancing
 *  of computation.
 *
 *  @param tg - (in) Task graph of a the mapped group of tasks
 *  @param netpm -- (in) Network performance model
 *  @param best_mg -- (out) Mapping solution generated by mapping heuristic 
 *
 */

int gs_smart_basic_map(gs_smart_tg * tg, gs_smart_netpm * netpm, gs_smart_mg ** _best_mg){
  gs_smart_mg * best_mg;
  gs_smart_tg_task_node * task_node;
  gs_smart_tg_remote_task_node * tg_rem_task_node;
  gs_server_t * server;
  int * mapping_vec;  
  int i, j, nb_rem_tasks, rem_task_cnt;
  int nbl_task_called;
  double this_time, best_time;
  int valid_assignment;
  int found_mapping;
  int nb_valid_mappings_timed;
  int valid_graph;
  int best_server_id=0;
  nb_rem_tasks=tg->nb_rem_nodes;
  best_mg = (gs_smart_mg * )calloc(1, sizeof(gs_smart_mg));
  mapping_vec = (int *)calloc(nb_rem_tasks, sizeof(int));
 
  rem_task_cnt=0;
  nbl_task_called=0;
  for(i=0;i<tg->nb_nodes;i++){
    task_node=tg->task_nodes[i];
    if(task_node->node_type==GS_SMART_TG_REM_TASK_NODE){
      tg_rem_task_node=task_node->tg_rem_task_node;

      nb_valid_mappings_timed=0;
      best_time=0.0;
      /*
       * find fastest smart server
       */
      for(j=0;j<netpm->nb_nodes;j++){
        server=netpm->netpm_nodes[j]->server;
        if(server->smart){
          if(gs_smart_check_if_valid_assignment(tg_rem_task_node,
                                   server, &valid_assignment)<0){
            ERRPRINTF("SMART : Error checking if a valid assignment\n");
            return -1;
          }
          
          if((valid_assignment) & (nb_valid_mappings_timed==0)){
              best_server_id=j;
              if(gs_smart_time_tg_task_node(tg_rem_task_node, server, &best_time)<0){
                ERRPRINTF("SMART : Error timing task graph task node\n");
                return -1;
              }
              nb_valid_mappings_timed++;
            }
            else if ((valid_assignment) & (nb_valid_mappings_timed>0)){
              if(gs_smart_time_tg_task_node(tg_rem_task_node, server, &this_time)<0){
                ERRPRINTF("SMART : Error timing task graph task node\n");
                return -1;
              }
              if(this_time<best_time){
                best_server_id=j;
                best_time=this_time;
              }
              nb_valid_mappings_timed++;
            }
            
          } 
        }
        
       /*
        * if no smart servers find fastest standard server
        */

        if(nb_valid_mappings_timed==0){
          found_mapping=0;
            for(j=0;j<netpm->nb_nodes;j++){
              server=netpm->netpm_nodes[j]->server;
              if(!server->smart){
                if(gs_smart_check_if_valid_assignment(tg_rem_task_node,
                                     server, &valid_assignment)<0){
                  ERRPRINTF("SMART : Error checking if a valid assignment\n");
                  return -1;
                }
                if((valid_assignment) & (nb_valid_mappings_timed==0)){
                  best_server_id=j;
                  if(gs_smart_time_tg_task_node(tg_rem_task_node, server, &best_time)<0){
                    ERRPRINTF("SMART : Error timing task graph task node\n");
                    return -1;
                  }
                  nb_valid_mappings_timed++;
                }
                else if ((valid_assignment) & (nb_valid_mappings_timed>0)){
                  if(gs_smart_time_tg_task_node(tg_rem_task_node, server, &this_time)<0){
                    ERRPRINTF("SMART : Error timing task graph task node\n");
                    return -1;
                  }
                  if(this_time<best_time){
                    best_server_id=j;
                    best_time=this_time;
                  }
                  nb_valid_mappings_timed++;
                }
              } 
            }
        } 
        if(nb_valid_mappings_timed==0){
          ERRPRINTF("SMART: Error no valid mappings found for task %d %s\n", i, tg_rem_task_node->problem->name);
          return -1;
        }


        /*
         * if its a blocking task reset the task workloads of all the servers
         */ 
        
        if(tg_rem_task_node->blocking==1){ 
          for(j=0;j<netpm->nb_nodes;j++){
            netpm->netpm_nodes[j]->server->task_workload=0;
          }
        }
        /*
         * if its a non-blocking task bump up the task workload of assigned server
         */ 
        else if(tg_rem_task_node->blocking==0){
          netpm->netpm_nodes[best_server_id]->server->task_workload =(server->task_workload+(( best_time*100000 )));
        }

      
      mapping_vec[rem_task_cnt]=best_server_id; 
      rem_task_cnt++;
    }

  }
  
  if(gs_smart_generate_mapping_graph(tg, netpm, mapping_vec, best_mg, &valid_graph)<0){
    ERRPRINTF("SMART: Error generating mapping graph\n");
    return -1;
  }
  if(gs_smart_mg_create_argument_file_names(best_mg)<0){
    ERRPRINTF("SMART : Error creating mapping graph argument file names\n");
    return -1;
  }


  if(!valid_graph){
    ERRPRINTF("SMART : The mapping graph generated is invalid!\n");
    return -1;
  }
  *_best_mg=best_mg;
  return 0;
}







/*
 *  This mapping heuristic randomly traverses the solution space and
 *  returns the fastest solution within that solution space.
 *  The number of solutions timed is max_solutions.
 *
 *  @param tg - (in) Task graph of a the mapped group of tasks
 *  @param netpm -- (in) Network performance model
 *  @param best_mg -- (out) Mapping solution generated by mapping heuristic 
 *
 */





int gs_smart_rwalk_map(gs_smart_tg * tg, gs_smart_netpm * netpm, gs_smart_mg ** _best_mg){

  gs_smart_mg * this_mg;
  gs_smart_mg * best_mg=NULL;
  int ** mapping_matrix;
  int * mapping_vec;  
  int nb_rem_tasks, nb_servers;
  int i, j;
  int total_permutations;
  double best_time=0.0;
  double this_time=0.0;
  int solution_num;
  int max_solutions=10000;
  int valid_graph=0;
  int sol_num=0;
  int num_valid_graphs=0;
  if(tg->nb_rem_nodes<=0){
    ERRPRINTF("SMART: Error doing exhaustive search, no remote nodes\n");
    return -1;
  }

  nb_servers=netpm->nb_nodes;
  nb_rem_tasks=tg->nb_rem_nodes;
  mapping_matrix = (int **)calloc( nb_servers, sizeof(int *));

  if(!mapping_matrix) return -1;
  
  for (i = 0; i <nb_servers; i++) {
    mapping_matrix[i] = (int *)calloc(nb_rem_tasks, sizeof(int));
    if(!mapping_matrix[i]) return -1;
  }

  for( i = 0; i < nb_servers; i++){
    for( j = 0; j < nb_rem_tasks; j++){
      mapping_matrix[i][j] = 0;
    }
  }

  mapping_vec = (int *)calloc(nb_rem_tasks, sizeof(int));
  if(!mapping_vec){
    ERRPRINTF("SMART: Error allocating to mapping vector\n");
    return -1;
  } 

  total_permutations=(int)(pow((double)nb_servers,(double)nb_rem_tasks));

  num_valid_graphs=0;
  while((sol_num<max_solutions) & (sol_num<total_permutations)){
    this_mg = (gs_smart_mg * )calloc(1, sizeof(gs_smart_mg));
    srand((unsigned)time(0)); 
    solution_num=(rand()%total_permutations);
    if(gs_smart_zero_mapping_matrix(mapping_matrix, nb_servers, nb_rem_tasks)<0){
      ERRPRINTF("SMART: Error resetting mapping matrix on mapping solution\n");
      return -1;
    }
    if(gs_smart_next_permutation(mapping_matrix, solution_num, nb_servers, nb_rem_tasks)<0){
      ERRPRINTF("SMART: Error shifting matrix to next permuation on mapping solution\n");
      return -1;
    }
    if(gs_smart_matrix_to_vec(mapping_matrix, nb_servers, nb_rem_tasks, mapping_vec)<0){
      ERRPRINTF("SMART: Error converting mapping matrix to a vector\n");
      return -1;
    }
    if(gs_smart_generate_mapping_graph(tg, netpm, mapping_vec, this_mg, &valid_graph)<0){
      ERRPRINTF("SMART: Error generating mapping graph\n");
      return -1;
    }
    if(!valid_graph){
      if(gs_smart_free_mg(this_mg)<0){
        ERRPRINTF("SMART : Error freeing mapping graph\n");
        return -1;
      }
    }
    else{

      this_time=this_mg->total_time;
      if(valid_graph){ 
        if(num_valid_graphs==0){
          best_time=this_time;
          best_mg=this_mg;
        }
        else{
          if(this_time<best_time){
            best_time=this_time;
            if(gs_smart_free_mg(best_mg)<0){
              ERRPRINTF("SMART : Error freeing mapping graph\n");
              return -1;
            }
            best_mg=this_mg;
          }
          
          else{
            if(gs_smart_free_mg(this_mg)<0){
              ERRPRINTF("SMART : Error freeing mapping graph\n");
              return -1;
            }
          }
        }
        num_valid_graphs++;
       }
     }
    sol_num++;
  }
  if(num_valid_graphs==0){
    ERRPRINTF("SMART: Could not find a valid graph\n");
    return -1;
  }
  if(gs_smart_mg_create_argument_file_names(best_mg)<0){
    ERRPRINTF("SMART : Error creating mapping graph argument file names\n");
    return -1;
  }
  

  *_best_mg=best_mg;

  return 0;

}








/*
 *  This mapping heuristic is a non-iterative heuristic.
 *  It randomly chooses one solution. This mapping solution
 *  should only be used for testing.
 *
 *  @param tg - (in) Task graph of a the mapped group of tasks
 *  @param netpm -- (in) Network performance model
 *  @param best_mg -- (out) Mapping solution generated by mapping heuristic 
 *
 */



int gs_smart_random_map(gs_smart_tg * tg, gs_smart_netpm * netpm, gs_smart_mg ** _best_mg){

  gs_smart_mg * this_mg;
  gs_smart_mg * best_mg=NULL;
  int ** mapping_matrix;
  int * mapping_vec;  
  int nb_rem_tasks, nb_servers;
  int i, j;
  int total_permutations;
  int solution_num;
  int max_solutions=10000;
  int valid_graph=0;
  int sol_num=0;

  this_mg = (gs_smart_mg * )calloc(1, sizeof(gs_smart_mg));
  if(tg->nb_rem_nodes<=0){
    ERRPRINTF("SMART: Error doing exhaustive search, no remote nodes\n");
    return -1;
  }

  nb_servers=netpm->nb_nodes;
  nb_rem_tasks=tg->nb_rem_nodes;
  mapping_matrix = (int **)calloc( nb_servers, sizeof(int *));

  if(!mapping_matrix) return -1;
  
  for (i = 0; i <nb_servers; i++) {
    mapping_matrix[i] = (int *)calloc(nb_rem_tasks, sizeof(int));
    if(!mapping_matrix[i]) return -1;
  }

  for( i = 0; i < nb_servers; i++){
    for( j = 0; j < nb_rem_tasks; j++){
      mapping_matrix[i][j] = 0;
    }
  }

  mapping_vec = (int *)calloc(nb_rem_tasks, sizeof(int));
  if(!mapping_vec){
    ERRPRINTF("SMART: Error allocating to mapping vector\n");
    return -1;
  } 

  total_permutations=(int)(pow((double)nb_servers,(double)nb_rem_tasks));

  while((!valid_graph) & (sol_num<max_solutions)){
    srand((unsigned)time(0)); 
    solution_num=(rand()%total_permutations);
    if(gs_smart_zero_mapping_matrix(mapping_matrix, nb_servers, nb_rem_tasks)<0){
      ERRPRINTF("SMART: Error resetting mapping matrix on mapping solution\n");
      return -1;
    }
    if(gs_smart_next_permutation(mapping_matrix, solution_num, nb_servers, nb_rem_tasks)<0){
      ERRPRINTF("SMART: Error shifting matrix to next permuation on mapping solution\n");
      return -1;
    }
    if(gs_smart_matrix_to_vec(mapping_matrix, nb_servers, nb_rem_tasks, mapping_vec)<0){
      ERRPRINTF("SMART: Error converting mapping matrix to a vector\n");
      return -1;
    }
    if(gs_smart_generate_mapping_graph(tg, netpm, mapping_vec, this_mg, &valid_graph)<0){
      ERRPRINTF("SMART: Error generating mapping graph\n");
      return -1;
    }
    if(!valid_graph){
/*
      if(gs_smart_free_mg(this_mg)<0){
        ERRPRINTF("SMART : Error freeing mapping graph\n");
        return -1;
      }
*/
    }
    sol_num++;
  }
  if(!valid_graph){
    ERRPRINTF("SMART: Could not find a valid graph\n");
    return -1;
  }
  if(gs_smart_mg_create_argument_file_names(this_mg)<0){
    ERRPRINTF("SMART : Error creating mapping graph argument file names\n");
    return -1;
  }
   best_mg=this_mg;
  *_best_mg=best_mg;

  return 0;

}




/*
 * This function sets the mapping heuristic which will be
 * invoked.
 *
 * @param _mapper_type -- the name of the mapper to invoke
 */


int gs_smart_set_mapper_type(char * _mapper_type){

  int i;
  if(!_mapper_type) return -1;
  for(i=0;i<nb_mappers;i++){
    if(strcmp(_mapper_type, mapping_names[i])==0){
      mapper_type=_mapper_type;
      return 0;
    }
  }
  return -1;
}




/*
 * This function gets the mapping heuristic which will be
 * invoked.
 *
 * @param _mapper_type -- the name of the mapper which will be invoked
 */


int gs_smart_get_mapper_type(char ** _mapper_type){
  if(!mapper_type) return -1;
  *_mapper_type=mapper_type;
  return 0;
}

/*
 * This function returns the number of implemented mapping heuristic.
 */

int gs_smart_get_nb_mappers(){
  return nb_mappers;
}


/*
 * This function returns the mapping heuristic name.
 *
 * @param number -- the position of the mapping heuristic in mapping_names array
 */
char * gs_smart_get_mapping_name(int number){
  return  mapping_names[number];
}


/*
 * This function checks if the called mapping heuristic exists 
 * and has been implemented.
 *
 * @param mapper_type -- the name of the mapping heuristic to check
 */


int gs_smart_check_if_mapper_exists(char * mapper_type){
  int i;
  for(i=0;i<nb_mappers;i++){
    if(strcmp(mapping_names[i], mapper_type)==0){
      return 0;
    }
  }
  return -1;
}


#endif
