/**
 * @file
 *
 * This program will print a list of servers.  The usage is:
 *    GS_config <agent name>
 *
 */
/* $Id: gs_config.c,v 1.12 2007/04/11 16:47:34 seymour Exp $ */
/* $UTK_Copyright: $ */

#include "utility.h"

#include <stdlib.h>
#include <string.h>

#include "comm_data.h"
#include "comm_basics.h"
#include "comm_encode.h"
#include "general.h"

/**
 * Entry point for GS_config.
 *
 * @param argc -- arg count
 * @param argv -- array of command line args
 *
 * @returns EXIT_SUCCESS on success, EXIT_FAILURE on error.
 */

int
main(int argc, char **argv)
{
  char *msg, dottedIP[20], proxy_dottedIP[20], cid_string[2 * CID_LEN + 1];
  int i, sock, num_servers, agentport, tag;
  gs_server_t **server_list;

  if(argc < 2) {
    fprintf(stderr,"Usage: GS_config <agent name>\n");
    exit(EXIT_FAILURE);
  }
  
  initialize_sockets();

  agentport = getenv_int("GRIDSOLVE_AGENT_PORT", GRIDSOLVE_AGENT_PORT_DEFAULT);
  if((sock = gs_connect_direct(argv[1], agentport)) == INVALID_SOCKET) {
    fprintf(stderr,"%s cannot be contacted\n", argv[1]);
    exit(EXIT_FAILURE);
  }

  if((gs_send_tag(sock, GS_PROT_SERVER_LIST) < 0) ||
     (gs_send_string(sock, VERSION) < 0)) {
    fprintf(stderr,"Cannot communicate with %s\n", argv[1]);
    fprintf(stderr,"Error sending tag GS_PROT_SERVER_LIST\n");
    exit(EXIT_FAILURE);
  }

  if(gs_recv_tag(sock, &tag) < 0) {
    fprintf(stderr,"Error communicating with agent.\n");
    exit(EXIT_FAILURE);
  }

  if(tag != GS_PROT_OK) {
    if(tag == GS_PROT_VERSION_MISMATCH)
      fprintf(stderr, "Error: Agent is an incompatible version\n");
    else
      fprintf(stderr, "Error: Agent refused with code %d\n", tag);
    exit(EXIT_FAILURE);
  }

  if(gs_recv_int(sock, &num_servers) < 0) {
    fprintf(stderr,"Error communicating with agent.\n");
    exit(EXIT_FAILURE);
  }

  printf("AGENT: %s [%d servers]\n", argv[1], num_servers);

  if(num_servers <= 0)
    exit(EXIT_SUCCESS);

  server_list = (gs_server_t **) CALLOC(num_servers, sizeof(gs_server_t *));
  if(!server_list) {
    fprintf(stderr,"Failed to allocate memory for the server list.\n");
    exit(EXIT_FAILURE);
  }

  for(i=0;i<num_servers;i++) {
    server_list[i] = (gs_server_t *) CALLOC(1,sizeof(gs_server_t));
    if(gs_recv_string(sock, &msg) < 0) {
      fprintf(stderr,"Error communicating with agent.\n");
      exit(EXIT_FAILURE);
    }
    if(gs_decode_server(msg, server_list[i]) < 0) {
      fprintf(stderr,"Error decoding server information.\n");
      exit(EXIT_FAILURE);
    }
    FREE(msg);
  }

  for(i=0;i<num_servers;i++) {
    proxy_cid_to_str(cid_string, server_list[i]->componentid);
    proxy_ip_to_str(server_list[i]->ipaddress, dottedIP);
    proxy_ip_to_str(server_list[i]->proxyip, proxy_dottedIP);
    if(server_list[i]->proxyip != 0)
      printf("SERVER: %s (%s:%d, proxy=%s:%d) [cid=%s]\n", server_list[i]->hostname, 
        dottedIP, server_list[i]->port, proxy_dottedIP, 
        server_list[i]->proxyport, cid_string);
    else
      printf("SERVER: %s (%s:%d) [cid=%s]\n", server_list[i]->hostname, 
        dottedIP, server_list[i]->port, cid_string);
  }

  cleanup_sockets();

  return 0;
}
