/**
 * @file
 *
 * This program will ask the server to terminate.  The usage is:
 *    GS_killserver <agent name> <server name)
 *
 */
/* $Id: gs_kill_server.c,v 1.24 2007/04/11 16:47:34 seymour Exp $ */
/* $UTK_Copyright: $ */

#include "utility.h"

#include <stdlib.h>
#include <string.h>

#include "comm_data.h"
#include "comm_basics.h"
#include "comm_encode.h"
#include "general.h"

char * gs_get_password(char *prompt);

#define DIE_BAD_USAGE() \
  do { \
    fprintf(stderr, "Usage: GS_killserver <agent name> <server name>\n"); \
    fprintf(stderr, "Usage: GS_killserver -c <agent name> <server component id>\n"); \
    exit(EXIT_FAILURE); \
  } while(0)

/**
 * Entry point for GS_killserver.
 *
 * @param argc -- arg count
 * @param argv -- array of command line args
 *
 * @returns EXIT_SUCCESS on success, EXIT_FAILURE on error.
 */

int
main(int argc, char **argv)
{
  int sock, server_sock, agentport, tag, by_cid;
  gs_server_t *gs_server = NULL;
  char *msg, *password, *agent_name, *server_name;

  if((argc < 3) || (argc > 4))
    DIE_BAD_USAGE();

  if(!strcmp(argv[1], "-c"))
    by_cid = 1;
  else
    by_cid = 0;

  if(argc == 3) {
    if(by_cid)
      DIE_BAD_USAGE();
    else {
      agent_name = argv[1];
      server_name = argv[2];
    }
  }
  else {
    if(by_cid) {
      agent_name = argv[2];
      server_name = (char *)malloc(strlen(argv[3]) + 2);
      if(!server_name) {
        fprintf(stderr, "failed to malloc memory\n");
        exit(EXIT_FAILURE);
      }
      sprintf(server_name, "-c%s", argv[3]);
    }
    else
      DIE_BAD_USAGE();
  }

  password = gs_get_password("Password: ");

  if(!password) {
    fprintf(stderr, "Error reading password.\n");
    exit(EXIT_FAILURE);
  }

  agentport = getenv_int("GRIDSOLVE_AGENT_PORT", GRIDSOLVE_AGENT_PORT_DEFAULT);
  if((sock = gs_connect_direct(agent_name, agentport)) == INVALID_SOCKET) {
    fprintf(stderr, "%s cannot be contacted\n", agent_name);
    exit(EXIT_FAILURE);
  }

  if((gs_send_tag(sock, GS_PROT_KILL_SERVER) < 0) ||
     (gs_send_string(sock, VERSION) < 0)) {
    fprintf(stderr, "Error communicating with agent.\n");
    exit(EXIT_FAILURE);
  }

  if(gs_recv_tag(sock, &tag) < 0) {
    fprintf(stderr, "Error communicating with agent.\n");
    exit(EXIT_FAILURE);
  }

  if(tag != GS_PROT_OK) {
    if(tag == GS_PROT_VERSION_MISMATCH)
      fprintf(stderr, "Error: Agent is an incompatible version\n");
    else
      fprintf(stderr, "Error: Agent refused with code %d\n", tag);
    exit(EXIT_FAILURE);
  }

  if(gs_send_string(sock, server_name) < 0) {
    fprintf(stderr, "Unsuccessful (Sending Server)\n");
    exit(EXIT_FAILURE);
  }

  if(gs_recv_tag(sock,&tag) < 0) {
    fprintf(stderr, "Unsuccessful (receiving response)\n");
    exit(EXIT_FAILURE);
  }

  if(tag != GS_PROT_OK) {
    if(tag == GS_PROT_UNKNOWN_SERVER) {
      if(by_cid) {
        fprintf(stderr, "No Server with component id '%s'.\n", server_name+2);
        fprintf(stderr, "Try specifying the component id exactly as it is ");
        fprintf(stderr, "displayed by GS_config\n");
      }
      else {
        fprintf(stderr, "No Server named '%s'.\n", server_name);
        fprintf(stderr, "Try specifying the hostname exactly as it is ");
        fprintf(stderr, "displayed by GS_config\n");
      }
    }
    else if(tag == GS_PROT_MULTIPLE_SERVER) {
      fprintf(stderr, "Multiple servers matched name '%s'.\n\n", server_name);
      fprintf(stderr, "Disambiguate by specifying the component id (cid) ");
      fprintf(stderr, "using\nthe -c flag to GS_killserver.  For example:\n\n");
      fprintf(stderr, "   GS_killserver -c agent_name 20206a66691be13f\n");
      fprintf(stderr, "\nGS_config will display the component ids.\n");
    }
    else
      fprintf(stderr, "Unspecified error killing server '%s'.\n", 
         by_cid ? (server_name + 2) : server_name);

    exit(EXIT_FAILURE);
  }

  gs_server = (gs_server_t *)calloc(1, sizeof(gs_server_t));
  if(!gs_server) {
    fprintf(stderr, "Unsuccessful (Allocating Server Memory)\n");
    exit(EXIT_FAILURE);
  }

  if(gs_recv_string(sock, &msg) < 0) {
    fprintf(stderr, "Unsuccessful receiving server structure.\n");
    exit(EXIT_FAILURE);
  }

  if(gs_decode_server(msg, gs_server) <0){
    fprintf(stderr, "Unsuccessful (Decoding Server)\n");
    exit(EXIT_FAILURE);
  }

  server_sock = gs_connect_to_host(gs_server->componentid, gs_server->ipaddress,
     gs_server->port, gs_server->proxyip, gs_server->proxyport);

  if(server_sock < 0) {
    fprintf(stderr, "Unsuccessful (connecting server)\n");
    gs_server_free(gs_server);
    exit(EXIT_FAILURE);
  }

  if((gs_send_tag(server_sock, GS_PROT_KILL_SERVER) < 0) ||
     (gs_send_string(server_sock, VERSION) < 0)) {
    fprintf(stderr, "Unsuccessful (sending tag to Server)\n");
    gs_server_free(gs_server);
    exit(EXIT_FAILURE);
  }

  if(gs_recv_tag(server_sock, &tag) < 0) {
    fprintf(stderr, "Error communicating with server.\n");
    exit(EXIT_FAILURE);
  }

  if(tag != GS_PROT_OK) {
    if(tag == GS_PROT_VERSION_MISMATCH)
      fprintf(stderr, "Error: Server is an incompatible version\n");
    else
      fprintf(stderr, "Error: Server refused with code %d\n", tag);
    exit(EXIT_FAILURE);
  }

  if(gs_send_string(server_sock, password) < 0) {
    fprintf(stderr, "Unsuccessful (Sending Password)\n");
    gs_server_free(gs_server);
    exit(EXIT_FAILURE);
  }

  if(gs_recv_tag(server_sock,&tag) < 0) {
    fprintf(stderr, "Unsuccessful (receiving response)\n");
    gs_server_free(gs_server);
    exit(EXIT_FAILURE);
  }

  if(tag == GS_PROT_OK) {
    printf("Password Accepted and Successfully Killed\n");

    if(gs_send_tag(sock, GS_PROT_OK) < 0) {
      fprintf(stderr, "Unsuccessful (Sending Confirmation to Agent)\n");
      gs_server_free(gs_server);
      exit(EXIT_FAILURE);
    }
  }
  else {
    printf("Password Not Accepted and Not Killed\n");

    if(gs_send_tag(sock, GS_PROT_ERROR) < 0) {
      fprintf(stderr, "Unsuccessful (Sending Confirmation to Agent)\n");
      gs_server_free(gs_server);
      exit(EXIT_FAILURE);
    }
  }

  gs_server_free(gs_server);

  exit(EXIT_SUCCESS);
}
